/*

    main.cpp: VUCAKO Bench (CPU / memory / cache benchmark).

    21.4.2001

    Copyright (C) 2001 by Josef Pelikan, MFF UK Prague
        http://sun3.ms.mff.cuni.cz/~pepca/

    VUCAKO Bench is Open Source software (see http://www.opensource.org/
    for description what Open Source is); you may distributed it and/or
    modify it under the terms of GNU General Public License as published
    by the Free Software Foundation; either version 2, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    A copy of the GNU General Public License is available as
    `/usr/doc/copyright/GPL' in the Debian GNU/Linux distribution or on
    the World Wide Web at http://www.gnu.org/copyleft/gpl.html.  You can
    also obtain it by writing to the Free Software Foundation, Inc., 59
    Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "pch.h"

#define VERSION     "1.004"

//-----------------------------------------------------------
//  live cycle:

#define LiveCycle(n,total)	printf("%4.0f %%\r",((n)*100.0)/(total))

//-----------------------------------------------------------
//  eratosthen:

#define PrimeOrder        5555555
#define PrimeAssert     100000000

int eratosthen ( int primeOrder, int primeAssert )
{
    startTimer();
    CBitMask sieve;
    sieve.assert(2);
    sieve.assert(primeAssert);
    int n = 0;
    int prime = 1;
    do {                                    // find the next prime
        do prime++;
        while ( sieve.getUnsafe(prime) );
        int k = prime;
        do {
            sieve.setUnsafe(k,true);
            k += prime;
            } while ( k <= primeAssert );
		if ( !(n & 0x7FFF) ) LiveCycle(n,primeOrder);
        } while ( ++n < primeOrder );
    int elapsed = elapsedTime();
    printf("The %d-th prime number is: %d, elapsed time: %.2f sec\n",
           primeOrder,prime,elapsed*0.001);
    return elapsed;
}

//-----------------------------------------------------------
//  BYTE bitmap:

#define ImageWidth1          4096
#define ImageHeight1         4096
#define ImageIter1           48

#define ImageWidth2          8192
#define ImageHeight2         8192
#define ImageIter2           12

BYTE *newImage ( int width, int height )
{
    width  &= -4;
    height &= -4;
    return new BYTE[width*height];
}

void transpose ( BYTE *dest, int width, int height, BYTE *src )
{
    width  &= -4;
    height &= -4;
    unsigned *dp1 = (unsigned*)dest;        // destination pointers
    dest += width;
    unsigned *dp2 = (unsigned*)dest;
    dest += width;
    unsigned *dp3 = (unsigned*)dest;
    dest += width;
    unsigned *dp4 = (unsigned*)dest;
    dest += width;
    unsigned *sp0 = (unsigned*)src;
    unsigned *sp;
    unsigned dact1, dact2, dact3, dact4, sact;
    int i, j;
    for ( i = 0; i < height; i += 4 ) {     // 4 destination rows
        sp = sp0;
        for ( j = 0; j < width; j += 4 ) {  // 4 destination columns
            sact   = *sp;
            dact1  =   sact        & 0xFF;
            dact2  =  (sact >>  8) & 0xFF;
            dact3  =  (sact >> 16) & 0xFF;
            dact4  =  (sact >> 24) & 0xFF;
            sp += (height >> 2);
            dact1 |= ( sact        & 0xFF) <<  8;
            dact2 |= ((sact >>  8) & 0xFF) <<  8;
            dact3 |= ((sact >> 16) & 0xFF) <<  8;
            dact4 |= ((sact >> 24) & 0xFF) <<  8;
            sp += (height >> 2);
            dact1 |= ( sact        & 0xFF) << 16;
            dact2 |= ((sact >>  8) & 0xFF) << 16;
            dact3 |= ((sact >> 16) & 0xFF) << 16;
            dact4 |= ((sact >> 24) & 0xFF) << 16;
            sp += (height >> 2);
            dact1 |= ( sact        & 0xFF) << 24;
            dact2 |= ((sact >>  8) & 0xFF) << 24;
            dact3 |= ((sact >> 16) & 0xFF) << 24;
            dact4 |= ((sact >> 24) & 0xFF) << 24;
            sp += (height >> 2);
            *dp1++ = dact1;
            *dp2++ = dact2;
            *dp3++ = dact3;
            *dp4++ = dact4;
            }
        sp0++;
        dp1 = (unsigned*)dest;
        dest += width;
        dp2 = (unsigned*)dest;
        dest += width;
        dp3 = (unsigned*)dest;
        dest += width;
        dp4 = (unsigned*)dest;
        dest += width;
        }
}

//-----------------------------------------------------------
//  arbitrage sequence lookup:

#define MaxN        16
#define ArbitrageN  12
#define MinBound    1.0

double m[MaxN][MaxN];
int n;

void initInstance ()
{
    n = ArbitrageN;
    int i;
    RandomJames rnd;
    for ( i = 0; i < n; i++ )
        rnd.normalNumbers(1.0,0.05,5,m[i],n);
}

int result[MaxN];
int used[MaxN];
int len;

double best;
int bestSeq[MaxN];
int bestLen;

void solve ( double mul, int i )
{
    int j;
    double act;
    if ( i + 1 == len )
        for ( j = 0; j < n; j++ ) {
            if ( !used[j] )
                if ( (act = mul * m[result[i-1]][j] * m[j][result[0]]) > best ) {
                    best = act;
                    bestLen = len;
                    memcpy(bestSeq,result,i*sizeof(int));
                    bestSeq[i] = j;
                    }
            }
    else
        for ( j = 0; j < n; j++ )
            if ( !used[j] ) {
                used[j] = 1;
                result[i] = j;
                solve(mul*m[result[i-1]][j],i+1);
                used[j] = 0;
                }
}

void solveInstance ()
{
    int i;
    memset(used,0,sizeof(used));
    best = MinBound;
    bestLen = 0;
    for ( len = 2; len < n; len++ ) {
        for ( i = 0; i < n; i++ ) {
            used[i] = 1;
            result[0] = i;
            solve(1.0,1);
            used[i] = 0;
            }
		LiveCycle(len-1,n-1);
		}
    if ( bestLen > 0 ) {
        printf("Arbitrage profit:%7.3f, seq:",best);
        for ( i = 0; i < bestLen; i++ )
            printf(" %d",bestSeq[i]+1);
        printf(" %d\n",bestSeq[0]+1);
        }
    else
        puts("No arbitrage sequence exists");
}

//-----------------------------------------------------------
//  needle throwing simulation:

#define NeedleMax   100000000

int needleSimulation ()
{
    startTimer();
    RandomJames rnd;
    unsigned hit = 0;
    unsigned i;
    double y, height, angle;
    for ( i = 0; i++ < NeedleMax; ) {
        y = rnd.uniformNumber() * 2.0;
        angle = rnd.uniformNumber() * 3.141592653589793;
        height = sin(angle);
        if ( y + height >= 2.0 || y - height < 0.0 ) hit++;
		if ( !(i & 0x1FFFFF) ) LiveCycle(i,NeedleMax);
        }
    int elapsed = elapsedTime();
    printf("Needle simulation: N = %dM, elapsed time: %.2f sec, ratio = %.9f\n",
           NeedleMax/1000000,elapsed*0.001,(hit+0.0)/NeedleMax);
    return elapsed;
}

//-----------------------------------------------------------
//  fast mem-copy:

#define MemCopyMax   (16<<20)
#define MemCopyIter  100

int memCopy ()
{
    startTimer();
    char *a = new char[MemCopyMax];
    char *b = new char[MemCopyMax];
    int i;
    for ( i = 0; i++ < MemCopyIter; ) {
        memcpy(b,a,MemCopyMax);
        memcpy(a,b,MemCopyMax);
		LiveCycle(i,MemCopyIter);
        }
    delete []a;
    delete []b;
    int elapsed = elapsedTime();
    printf("Fast mem-copy: array size = %d MB, iterations = %d, elapsed time: %.2f sec\n",
           MemCopyMax>>20,MemCopyIter,elapsed*0.001);
    return elapsed;
}

//-----------------------------------------------------------
//  Monte-Carlo form-factor between two identical rectangles:

#define FFIter  50000000

int formFactor ()
{
    startTimer();
    int i;
    unsigned fullHits = 0;
    unsigned halfCount = 0;
    unsigned halfHits = 0;
    double x0, y0;                          // ray origin
    double x1, y1, z1;                      // ray direction
    double x2, y2;
    double tmp;
    bool half;
    RandomJames rnd;
    for ( i = 0; i++ < FFIter; ) {
        x0 = rnd.uniformNumber();
        y0 = rnd.uniformNumber();
        if ( (half = x0 < 0.5 && y0 < 0.5) ) halfCount++;
        do {
            x1 = rnd.uniformNumber() * 2.0 - 1.0;
            y1 = rnd.uniformNumber() * 2.0 - 1.0;
            tmp = x1 * x1 + y1 * y1;
            } while ( tmp >= 1.0 );
        z1 = sqrt( 1.0 - tmp );
        if ( z1 > 0.01 ) {                  // there is a chance
            tmp = 1.0 / z1;
            x2 = x0 + tmp * x1;
            y2 = y0 + tmp * y1;
            if ( x2 >= 0.0 && y2 >= 0.0 )
                if ( x2 < 1.0 && y2 < 1.0 ) {
                    fullHits++;
                    if ( half && x2 < 0.5 && y2 < 0.5 ) halfHits++;
                    }
            }
		if ( !(i & 0xFFFFF) ) LiveCycle(i,FFIter);
        }
    int elapsed = elapsedTime();
    printf("Form-factor: iter = %dM, time: %.2f sec, Ffull =%10.7f, Fhalf =%10.7f\n",
           FFIter/1000000,elapsed*0.001,(fullHits+0.0)/FFIter,(halfHits+0.0)/halfCount);
    return elapsed;
}

//-----------------------------------------------------------
//  Natural merge-sort of huge disk files:

#define SortSize        (1<<23)
#define BucketSize      1024
#define TestFileName    "sort%d.bin"

bool checkFile ( FileOfDouble &f, int &size )
{
    double prev, act;
    f.seek(0);
    act = f.get();
    size = 1;
    while ( true ) {
        prev = act;
        act = f.get();
        if ( act < 0.0 ) return true;
        size++;
        if ( act < prev ) return false;
        }
}

void initialSplit ( FileOfDouble &from, FileOfDouble &to1, int &seg1, FileOfDouble &to2, int &seg2 )
{
    to1.rewrite();
    seg1 = 0;
    to2.rewrite();
    seg2 = 0;
    from.seek(0);
    double prev, act;
    act = from.get();
    if ( act < 0.0 ) return;
    bool one = true;
    to1.put(act);
    seg1 = 1;

    do {
        prev = act;
        act = from.get();
        if ( act < 0.0 ) return;
        if ( act < prev ) {
            one = !one;
            if ( one ) seg1++;
            else       seg2++;
            }
        if ( one ) to1.put(act);
        else       to2.put(act);
        } while ( true );
}

void mergeAndSplit ( FileOfDouble &from1, FileOfDouble &from2, 
                     FileOfDouble &to1, int &seg1, FileOfDouble &to2, int &seg2 )
{
    to1.rewrite();
    seg1 = 0;
    to2.rewrite();
    seg2 = 0;
    from1.seek(0);
    from2.seek(0);
    double prev1, act1;
    double prev2, act2;
    prev1 = from1.get();
    act1 = from1.get();
    prev2 = from2.get();
    act2 = from2.get();
    bool one = true;
    seg1 = 1;

    do {
            // min { prev1, prev2 } should be written out
        if ( prev1 < prev2 ) {
            if ( one ) to1.put(prev1);
            else       to2.put(prev1);
            if ( prev1 > act1 ) {           // copy the rest of from2 segment
                // prev1 > act1, prev1 was processed
                do {
                    if ( one ) to1.put(prev2);
                    else       to2.put(prev2);
                    if ( act2 < prev2 ) break;
                    prev2 = act2;
                    act2 = from2.get();
                    } while ( true );
                // prev2 > act2, prev2 was processed
                if ( act1 < 0.0 || act2 < 0.0 ) // from1 or from2 were exhausted
                    break;
                one = !one;
                if ( one ) seg1++;
                else       seg2++;
                prev2 = act2;
                act2 = from2.get();
                }
            prev1 = act1;
            act1 = from1.get();
            }
        else {                              // prev2 < prev1
            if ( one ) to1.put(prev2);
            else       to2.put(prev2);
            if ( prev2 > act2 ) {           // copy the rest of from1 segment
                // prev2 > act2, prev2 was processed
                do {
                    if ( one ) to1.put(prev1);
                    else       to2.put(prev1);
                    if ( act1 < prev1 ) break;
                    prev1 = act1;
                    act1 = from1.get();
                    } while ( true );
                // prev1 > act1, prev1 was processed
                if ( act1 < 0.0 || act2 < 0.0 ) // from1 or from2 were exhausted
                    break;
                one = !one;
                if ( one ) seg1++;
                else       seg2++;
                prev1 = act1;
                act1 = from1.get();
                }
            prev2 = act2;
            act2 = from2.get();
            }
        } while ( true );

    while ( act1 >= 0.0 ) {                 // from1 wasn't exhausted
        one = seg1 < seg2;
        if ( one ) {                        // put the 1st item
            seg1++;
            to1.put(act1);
            }
        else {
            seg2++;
            to2.put(act1);
            }
        do {
            prev1 = act1;
            act1 = from1.get();
            if ( prev1 > act1 ) break;
            if ( one ) to1.put(act1);
            else       to2.put(act1);
            } while ( true );
        }

    if ( act2 >= 0.0 ) {                    // from2 wasn't exhausted
        one = seg1 < seg2;
        if ( one ) {                        // put the 1st item
            seg1++;
            to1.put(act2);
            }
        else {
            seg2++;
            to2.put(act2);
            }
        do {
            prev2 = act2;
            act2 = from2.get();
            if ( prev2 > act2 ) break;
            if ( one ) to1.put(act2);
            else       to2.put(act2);
            } while ( true );
        }
}

int CDECL compare ( const void *a, const void *b )
{
    if ( *((double*)a) < *((double*)b) ) return -1;
    return( *((double*)a) == *((double*)b) ? 0 : 1 );
}

int diskSort ()
{
    fprintf(stderr,"init: random generator, bucket-sort\r");
    startTimer();
    char fn1[32], fn2[32], fn3[32], fn4[32];
    sprintf(fn1,TestFileName,1);
    sprintf(fn2,TestFileName,2);
    sprintf(fn3,TestFileName,3);
    sprintf(fn4,TestFileName,4);
    RandomJames rnd;
    FileOfDouble f1(fn1);
    FileOfDouble f2(fn2);
    FileOfDouble f3(fn3);
    FileOfDouble f4(fn4);
    int seg1, seg2, seg3, seg4;
    seg3 = seg4 = 0;
    int i, j;
        // file init:
    for ( i = 0; i++ < SortSize/BucketSize; ) {
        double bucket[BucketSize];
        for ( j = 0; j < BucketSize; j++ )
            bucket[j] = rnd.uniformNumber();
        qsort(bucket,BucketSize,sizeof(double),compare);
        for ( j = 0; j < BucketSize; j++ )
            f3.put(bucket[j]);
        }
        // merge-sort:
    initialSplit(f3,f1,seg1,f2,seg2);
    bool ok;
    do {
        printf("%17d %17d\r",seg1,seg2);
        if ( seg2 == 0 ) {
            ok = checkFile(f1,i);
            break;
            }
        if ( seg1 == 0 ) {
            ok = checkFile(f2,i);
            break;
            }
        mergeAndSplit(f1,f2,f3,seg3,f4,seg4);
        printf("%17d %17d\r",seg3,seg4);
        if ( seg4 == 0 ) {
            ok = checkFile(f3,i);
            break;
            }
        if ( seg3 == 0 ) {
            ok = checkFile(f4,i);
            break;
            }
        mergeAndSplit(f3,f4,f1,seg1,f2,seg2);
        } while ( true );
        // the right order was checked
    f1.close();
    f2.close();
    f3.close();
    f4.close();
    remove(fn1);
    remove(fn2);
    remove(fn3);
    remove(fn4);
    int elapsed = elapsedTime();
    printf("Merge-sort on disk: %s, file size: %d MB, elapsed time: %.2f sec\n",
           (ok && i == SortSize) ? "OK":"fail",(SortSize*sizeof(double))>>20,elapsed*0.001);
    return elapsed;
}

//-----------------------------------------------------------
//  Wavelet transform (memory cache test):

#define WaveMaxSize     (1<<22)
#define WaveMinSize     (1<<11)
#define WaveIter        20

void liftMemory ( int n0, int32 *s0, int32 *s1, int32 *d1 )
    // forward TS-transform
{
    int n1 = n0 >> 1;
    int32 *ps0, *ps1, *pd1;
    int i;
        // split:
    ps0 = s0;
    ps1 = s1; pd1 = d1;
    for ( i = 0; i++ < n1; ) {
        *ps1++ = *ps0++;
        *pd1++ = *ps0++;
        }
        // S-transform:
    ps1 = s1; pd1 = d1;
    for ( i = 0; i++ < n1; ) {
        *pd1 -= *ps1;
        *ps1++ += (*pd1++ >> 1);
        }
        // predict:
    ps1 = s1; pd1 = d1;
    ps1[n1] = ps1[n1-1];                    // boundary
    ps1[-1] = ps1[0];
    for ( i = 0; i++ < n1; ) {
        *pd1++ += ((ps1[-1] - ps1[1] + 2) >> 2);
        ps1++;
        }
}

void unliftMemory ( int n0, int32 *s0, int32 *s1, int32 *d1 )
    // inverse TS-transform
{
    int n1 = n0 >> 1;
    int32 *ps0, *ps1, *pd1;
    int i;
        // predict*:
    ps1 = s1; pd1 = d1;
    ps1[n1] = ps1[n1-1];                    // boundary
    ps1[-1] = ps1[0];
    for ( i = 0; i++ < n1; ) {
        *pd1++ -= ((ps1[-1] - ps1[1] + 2) >> 2);
        ps1++;
        }
        // S-transform*:
    ps1 = s1; pd1 = d1;
    for ( i = 0; i++ < n1; ) {
        *ps1 -= (*pd1 >> 1);
        *pd1++ += *ps1++;
        }
        // merge:
    ps0 = s0;
    ps1 = s1; pd1 = d1;
    for ( i = 0; i++ < n1; ) {
        *ps0++ = *ps1++;
        *ps0++ = *pd1++;
        }
}

int waveletMemory ()
{
    startTimer();
    int32 *a = new int32[WaveMaxSize+2];
    int total = WaveMaxSize;
    int iter = WaveIter;
    int i;
    int el = 0, elOld;
    do {
        int32 *s1 = a + (total >> 1) + 1;
        int32 *d1 = s1 + (total >> 2) + 1;
        for ( i = 0; i++ < iter; ) {
            liftMemory(total>>1,a,s1,d1);
            unliftMemory(total>>1,a,s1,d1);
            }
        elOld = el;
        el = elapsedTime();
        printf("%6.2f",(el-elOld)*0.001);
        iter += iter;
        } while ( (total /= 2) >= WaveMinSize );
    delete []a;
    int elapsed = elapsedTime();
    printf("\nWavelet T-S transform: arr: %dMB-%dKB, iter: %d-%d, elapsed time: %.2f sec\n",
           (WaveMaxSize*sizeof(int32))>>20,(WaveMinSize*sizeof(int32))>>10,WaveIter,iter>>1,elapsed*0.001);
    return elapsed;
}

//-----------------------------------------------------------
//  Wavelet transform (disk cache test):

#define S0FileName      "s0.bin"
#define S1FileName      "s1.bin"
#define D1FileName      "d1.bin"
#define WaveDMaxSize    (1<<24)
#define WaveDMinSize    (1<<17)
#define WaveDIter       1

void liftDisk ( int n0, FileOfDouble &s0, FileOfDouble &s1, FileOfDouble &d1 )
    // forward TS-transform
{
    int n1 = n0 >> 1;
    int i;
        // split + S-transform + prediction:
    s0.seek(0);
    s1.seek(0);
    d1.seek(0);
    double s[3], d[3];
    s[1] = s0.get();
    d[1] = s0.get();
    d[1] -= s[1];
    s[1] += d[1] * 0.5;
    s[0] = 0.0;
    for ( i = 1; i++ < n1; ) {
        s[2] = s0.get();
        d[2] = s0.get();
        d[2] -= s[2];
        s[2] += d[2] * 0.5;
        d[1] += (s[2] + s[0]) * 0.25;
        s1.put(s[1]);
        d1.put(d[1]);
        s[0] = s[1];
        s[1] = s[2];
        d[1] = d[2];
        }
    s[2] = 0.0;
    d[1] += (s[2] + s[0]) * 0.25;
    s1.put(s[1]);
    d1.put(d[1]);
}

void unliftDisk ( int n0, FileOfDouble &s0, FileOfDouble &s1, FileOfDouble &d1 )
    // inverse TS-transform
{
    int n1 = n0 >> 1;
    int i;
        // predict* + S-transform* + merge:
    s0.seek(0);
    s1.seek(0);
    d1.seek(0);
    double s[3], d[3];
    s[0] = 0.0;
    s[1] = s1.get();
    d[1] = d1.get();
    for ( i = 1; i++ < n1; ) {
        s[2] = s1.get();
        d[2] = d1.get();
        d[1] -= (s[2] + s[0]) * 0.25;
        double s1act = s[1] - d[1] * 0.5;
        d[1] += s1act;
        s0.put(s1act);
        s0.put(d[1]);
        s[0] = s[1];
        s[1] = s[2];
        d[1] = d[2];
        }
    s[2] = 0.0;
    d[1] -= (s[2] + s[0]) * 0.25;
    s[1] -= d[1] * 0.5;
    d[1] += s[1];
    s0.put(s[1]);
    s0.put(d[1]);
}

int waveletDisk ()
{
    FileOfDouble s0(S0FileName);
    FileOfDouble s1(S1FileName);
    FileOfDouble d1(D1FileName);
    int total = WaveDMaxSize;
    int iter = WaveDIter;
    int i;
    RandomJames rnd;
    startTimer();
    do {
        s0.rewrite();
        s1.rewrite();
        d1.rewrite();
        i = total >> 1;
        do
            s0.put(rnd.uniformNumber());
        while ( --i );
        int elOld = elapsedTime();
        for ( i = 0; i++ < iter; ) {
            liftDisk(total>>1,s0,s1,d1);
            unliftDisk(total>>1,s0,s1,d1);
            }
        int el = elapsedTime();
        printf("%6.2f",(el-elOld)*0.001);
        iter += iter;
        } while ( (total /= 2) >= WaveDMinSize );
    int elapsed = elapsedTime();
    s0.close();
    s1.close();
    d1.close();
    remove(S0FileName);
    remove(S1FileName);
    remove(D1FileName);
    printf("\nWavelet T-S transform on disk: arr: %dMB-%dMB, iter: %d-%d, time: %.2f sec\n",
           (WaveDMaxSize*sizeof(double))>>20,(WaveDMinSize*sizeof(double))>>20,WaveDIter,iter>>1,elapsed*0.001);
    return elapsed;
}

//-----------------------------------------------------------
//  Adaptive K-D tree on disk:

#define TreeInsertions      50000
#define TreeWalk           100000
#define TreeOperations     (TreeInsertions+TreeWalk)
#define TreeCache            2048
#define TreeXmax            10240
#define TreeYmax            10240
#define TreeWalkAmp          1024
#define TreeXhalf           (TreeXmax>>1)
#define TreeFile            "akd-tree.bin"

int akdTree ()
{
    startTimer();
    RandomJames rnd;
    AKDtree tree(TreeFile,TreeXhalf,TreeCache);
    int i;
    unsigned stat[5];
    for ( i = 0; i++ < TreeInsertions; ) {
        int32 x = (int32)(rnd.uniformNumber() * TreeXmax);
        int32 y = (int32)(rnd.uniformNumber() * TreeYmax);
        tree.insert(x,y);
        if ( !(i & 0x1FFF) ) {
            tree.getStat(stat);
            printf("Op: %dK, rd: %uK, wr: %uK, fl: %uK, leaf: %u\r",
                   i/1000,stat[0]/1000,stat[1]/1000,stat[2]/1000,stat[4]);
            }
        }
    double x = TreeXmax * 0.5;
    double y = TreeYmax * 0.5;
    while ( i++ < TreeOperations ) {
        double dx = rnd.uniformNumber() * TreeWalkAmp - (TreeWalkAmp/2);
        double dy = rnd.uniformNumber() * TreeWalkAmp - (TreeWalkAmp/2);
        if ( (x += dx) < 1.0 ) x += TreeXmax;
        if ( x >= TreeXmax-1.0 ) x -= TreeXmax;
        if ( (y += dy) < 1.0 ) y += TreeYmax;
        if ( y >= TreeYmax-1.0 ) y -= TreeYmax;
        if ( rnd.uniformNumber() < 1.0/3.0 ) tree.insert((int)x,(int)y);
        else                                 tree.search((int)x,(int)y);
        if ( !(i & 0x1FFF) ) {
            tree.getStat(stat);
            printf("Op: %dK, rd: %uK, wr: %uK, fl: %uK, leaf: %u\r",
                   i/1000,stat[0]/1000,stat[1]/1000,stat[2]/1000,stat[4]);
            }
        }
    tree.getStat(stat);
    tree.close();
    remove(TreeFile);
    int elapsed = elapsedTime();
    printf("Adaptive K-D tree on disk: operations: %dK, elapsed time: %.2f sec\n",
           TreeOperations/1000,elapsed*0.001);
    printf("Block reads: %uK, writes: %uK, flushes: %uK, leaf nodes: %u\n",
           stat[0]/1000,stat[1]/1000,stat[2]/1000,stat[4]);
    return elapsed;
}

//-----------------------------------------------------------
//  data persistence:

#define DataFile        "results.txt"
#define InitData        32
#define MaxProcName     128
#define CPUFile         "cpu.txt"
#define UndefinedCPU    "undefined"
#define UndefinedClock  1

struct DataItem {
    char procName[MaxProcName];
    int clock;
    int stat;
    int time;
    };

DataItem *di = NULL;
int diLen = 0;
int diAlloc = 0;

const char *actCpu = UndefinedCPU;
int actClock = 0;
char cpuBuf[MaxProcName];

void assertData ( int i )
{
    if ( i <= diAlloc ) return;
    i = (i + InitData-1) & -InitData;
    DataItem *ndi = new DataItem[i];
    if ( diAlloc ) {
        memcpy(ndi,di,diAlloc*sizeof(DataItem));
	    delete []di;
		}
    di = ndi;
    diAlloc = i;
}

bool readData ()
    // <test#> <time-in-millisec> <clock-MHz> <CPU-name>
{
    char buf[MaxProcName<<1];
    diLen = 0;
    FILE *f = fopen(DataFile,"rt");
    if ( !f ) return false;
    do {
        if ( !fgets(buf,MaxProcName<<1,f) ) break;
        char *ptr = buf;
        int stat = 0;
        int read;
        if ( sscanf(ptr,"%d%n",&stat,&read) < 1 || stat < 1 ) break;
        ptr += read;
        while ( buf[strlen(buf)-1] == '\n' || isspace(buf[strlen(buf)-1]) )
            buf[strlen(buf)-1] = (char)0;
        int ti, cl;
        if ( sscanf(ptr,"%d%d%n",&ti,&cl,&read) < 2 ) break;
        ptr += read;
        assertData(diLen+1);
        di[diLen].stat = stat;
        di[diLen].time = ti;
        di[diLen].clock = cl;
        while ( isspace(*ptr) ) ptr++;
        strncpy(di[diLen].procName,ptr,MaxProcName);
        di[diLen].procName[MaxProcName-1] = (char)0;
        diLen++;
        } while ( true );
    fclose(f);
    return true;
}

static bool verbose = true;
static bool changed = false;

bool writeData ()
{
    if ( !changed ) return true;
    FILE *f = fopen(DataFile,"wt");
    if ( !f ) return false;
    int i;
    for ( i = 0; i < diLen; i++ )
        fprintf(f,"%2d %7d %4d %s\n",di[i].stat,di[i].time,di[i].clock,di[i].procName);
    fclose(f);
    return true;
}

void newResult ( int stat, int elapsed )
{
    bool insert = true;
    printf("%-44s%7.2f sec (%6.2f%%), ref:%8.3f\n",
           actCpu,elapsed*0.001,100.0,
           elapsed*1.e-6*actClock);
        // print comparisons:
	char buf[48];
    int i;
    for ( i = 0; i < diLen; i++ )
        if ( di[i].stat == stat )
            if ( strcmp(di[i].procName,actCpu) || di[i].clock != actClock ) {
                if ( verbose ) {            // compare it
					strncpy(buf,di[i].procName,44);
					buf[44] = (char)0;
                    printf("%-44s%7.2f sec (%6.2f%%), ref:%8.3f\n",
                           buf,di[i].time*0.001,(elapsed*100.0)/di[i].time,
                           di[i].time*1.e-6*di[i].clock);
					}
                }
            else {                          // improve?
                if ( elapsed < di[i].time ) {
                    di[i].time = elapsed;
                    changed = true;
                    }
                insert = false;
                }
        // insert the new result:
    if ( insert ) {
        changed = true;
        assertData(diLen+1);
        di[diLen].stat = stat;
        di[diLen].time = elapsed;
        di[diLen].clock = actClock;
        strcpy(di[diLen++].procName,actCpu);
        }
}

void listResult ( int stat )
{
    printf("\n  ");
    switch ( stat ) {
        case  1:
            printf("Eratosthen (%d-th prime number):\n",PrimeOrder);
            break;
        case  2:
            printf("Transposition of %dMB image:\n",(ImageWidth1*ImageHeight1)>>20);
            break;
        case  3:
            printf("Transposition of %dMB image:\n",(ImageWidth2*ImageHeight2)>>20);
            break;
        case  4:
            printf("Arbitrage sequence lookup (double[%d][%d] matrix):\n",ArbitrageN,ArbitrageN);
            break;
        case  5:
            printf("Needle-throwing simulation (Monte-Carlo, %dM iterations):\n",NeedleMax/1000000);
            break;
        case  6:
            printf("Fast memory-copy (two %d MB arrays, %d iterations):\n",MemCopyMax>>20,MemCopyIter);
            break;
        case  7:
            printf("Monte-Carlo form-factor (%dM iterations):\n",FFIter/1000000);
            break;
        case  8:
            printf("Merge-sort on disk (double[%dM] array):\n",SortSize>>20);
            break;
        case  9:
            printf("Wavelet (T-S) transform in memory (int32[%dM-%dK] arrays):\n",WaveMaxSize>>20,WaveMinSize>>10);
            break;
        case 10:
            printf("Wavelet (T-S) transform on disk (double[%dM-%dK] arrays):\n",WaveDMaxSize>>20,WaveDMinSize>>10);
            break;
        case 11:
            printf("Adaptive K-D tree on disk (%dK operations):\n",TreeOperations/1000);
            break;
        }
        // print comparisons:
	char buf[56];
    int i;
    for ( i = 0; i < diLen; i++ )
        if ( di[i].stat == stat ) {
			strncpy(buf,di[i].procName,54);
			buf[54] = (char)0;
            printf("%-54s%7.2f sec, ref:%8.3f\n",
                   buf,di[i].time*0.001,di[i].time*1.e-6*di[i].clock);
			}
}

//-----------------------------------------------------------
//  main routine:

#define MaxTestMask     0x07FF
#define DefaultTestMask 0x07FB

void printUsage ()
{
    fprintf(stderr,
            "Usage: bench [-h] [-l] [-b] [<mask> [<cpu-clock> [\"cpu-description\"]]]\n"
            "    -h     print help message (this text)\n"
            "    -l     don't run any tests, print result table(s) only\n"
            "    -b     brief mode (don't print old results)\n"
            "    <mask> is bit-mask (hexadecimal number 0xNNN can be used):\n"
            "\t0x001 .. Eratosthen (%d-th prime number)\n"
            "\t0x002 .. Transposition of %dMB image (uses two buffers!)\n"
            "\t0x004 .. Transposition of %dMB image (uses two buffers!)\n"
            "\t0x008 .. Arbitrage sequence lookup (double[%d][%d] matrix)\n"
            "\t0x010 .. Needle-throwing simulation (Monte-Carlo, %dM iterations)\n"
            "\t0x020 .. Fast memory-copy (two %d MB arrays, %d iterations)\n"
            "\t0x040 .. Monte-Carlo form-factor (%dM iterations)\n"
            "\t0x080 .. Merge-sort on disk (double[%dM] array)\n"
            "\t0x100 .. Wavelet (T-S) transform in memory (int32[%dM-%dK] arrays)\n"
            "\t0x200 .. Wavelet (T-S) transform on disk (double[%dM-%dK] arrays)\n"
            "\t0x400 .. Adaptive K-D tree on disk (%dK operations)\n"
            "    <cpu-clock> is CPU clock in MHz (see the 2nd column of \"%s\")\n"
            "    \"cpu-description\" is arbitrary string (abbreviations are preferred,\n"
            "                      see \"%s\" ignoring first two columns)\n"
            "    if <cpu-clock> and \"cpu-description\" are not specified, program will\n"
            "      read the first line of \"%s\"\n"
            ,PrimeOrder,(ImageWidth1*ImageHeight1)>>20,
            (ImageWidth2*ImageHeight2)>>20,ArbitrageN,ArbitrageN,NeedleMax/1000000,
            MemCopyMax>>20,MemCopyIter,FFIter/1000000,SortSize>>20,
            WaveMaxSize>>20,WaveMinSize>>10,WaveDMaxSize>>20,WaveDMinSize>>10,TreeOperations/1000,
            CPUFile,CPUFile,CPUFile);
}

int CDECL main ( int argc, char **argv )
{
    printf("VUCAKO Bench ver %s (c) 2001 Josef.Pelikan@mff.cuni.cz (Open Source)\n"
        "  Please send result file (%s) back to author!  Help: \"bench -h\"\n",VERSION,DataFile);

    unsigned mask = DefaultTestMask;
    bool listing = false;
    int arg = 1;
    while ( arg < argc && argv[arg][0] == '-' ) { // read one "-*"  argument
        if ( argv[arg][1] == 'h' ) {        // help
            printUsage();
            return 1;
            }
        else
        if ( argv[arg][1] == 'l' ) {        // listing
            listing = true;
            }
        else
        if ( argv[arg][1] == 'b' ) {        // brief mode
            verbose = false;
            }
        else
            fprintf(stderr,"Unknown argument \"%s\" (will be ignored), type \"bench -h\" for help!\n",
                    argv[arg]);
        arg++;
        }
    bool wasCpu = false;
    if ( arg < argc ) {
        mask = (unsigned)strtol(argv[arg++],NULL,0) & MaxTestMask;
        if ( arg < argc ) {                 // CPU is specified in the command-line
            wasCpu = true;
            actClock = atoi(argv[arg++]);
            if ( arg < argc ) actCpu = argv[arg++];
            }
        }

    if ( !wasCpu ) {                        // read CPU-info from CPUFile
        FILE *f = fopen(CPUFile,"rt");
        if ( f ) {
            fscanf(f,"%*s%d",&actClock);    // skip the e-mail
            fgets(cpuBuf,MaxProcName,f);
            while ( cpuBuf[strlen(cpuBuf)-1] == '\n' ) cpuBuf[strlen(cpuBuf)-1] = (char)0;
            actCpu = cpuBuf;
            while ( isspace(*actCpu) ) actCpu++;
            fclose(f);
            }
        else {
            fprintf(stderr,"CPU type & frequency were not specified in the command line,\n"
                           "CPU file \"%s\" was not found => using default values [\"%s\",%d]!\n",
                    CPUFile,actCpu,actClock);
            }
        }

    if ( !readData() )
        fprintf(stderr,"Error reading old result file \"%s\"!\n",DataFile);

    if ( listing ) {                        // print only old results
        arg = 0;
        while ( (1 << arg) & MaxTestMask ) {
			if ( (1 << arg) & mask )
				listResult(arg);
			arg++;
			}
        return 0;
        }

    printf("Running tests \"%04X\" on \"%s\" (%d MHz):\n",mask,actCpu,actClock);
    int elapsed;

        // 1 - eratosthen:
    if ( mask & 0x0001 ) {
        elapsed = eratosthen(PrimeOrder,PrimeAssert);
        newResult(1,elapsed);
        }

        // image transposition:
    if ( mask & 0x0002 ) {                  // 2 - small image
        startTimer();
        BYTE *im1 = newImage(ImageWidth1,ImageHeight1);
        BYTE *im2 = newImage(ImageHeight1,ImageWidth1);
        int i;
        for ( i = 0; i++ < ImageIter1; ) {
            transpose(im2,ImageHeight1,ImageWidth1,im1);
            transpose(im1,ImageWidth1,ImageHeight1,im2);
			LiveCycle(i,ImageIter1);
            }
        delete []im1;
        delete []im2;
        elapsed = elapsedTime();
        printf("Transpose the [%d,%d] image, elapsed time: %.2f sec\n",
               ImageWidth1,ImageHeight1,elapsed*0.001);
        newResult(2,elapsed);
        }

    if ( mask & 0x0004 ) {                  // 3 - bigger image
        startTimer();
        BYTE *im1 = newImage(ImageWidth2,ImageHeight2);
        BYTE *im2 = newImage(ImageHeight2,ImageWidth2);
        int i;
        for ( i = 0; i++ < ImageIter2; ) {
            transpose(im2,ImageHeight2,ImageWidth2,im1);
            transpose(im1,ImageWidth2,ImageHeight2,im2);
			LiveCycle(i,ImageIter2);
            }
        delete []im1;
        delete []im2;
        elapsed = elapsedTime();
        printf("Transpose the [%d,%d] image, elapsed time: %.2f sec\n",
               ImageWidth2,ImageHeight2,elapsed*0.001);
        newResult(3,elapsed);
        }

        // 4 - arbitrage sequence:
    if ( mask & 0x0008 ) {
        startTimer();
        initInstance();
        solveInstance();
        elapsed = elapsedTime();
        newResult(4,elapsed);
        }

        // 5 - needle throwing simulation:
    if ( mask & 0x0010 ) {
        elapsed = needleSimulation();
        newResult(5,elapsed);
        }

        // 6 - fast mem-copy:
    if ( mask & 0x0020 ) {
        elapsed = memCopy();
        newResult(6,elapsed);
        }

        // 7 - Monte-Carlo form-factor:
    if ( mask & 0x0040 ) {
        elapsed = formFactor();
        newResult(7,elapsed);
        }

        // 8 - Merge-sort on disk:
    if ( mask & 0x0080 ) {
        elapsed = diskSort();
        newResult(8,elapsed);
        }

        // 9 - T-S wavelet transform in memory:
    if ( mask & 0x0100 ) {
        elapsed = waveletMemory();
        newResult(9,elapsed);
        }

        // 10 - T-S wavelet transform on disk:
    if ( mask & 0x0200 ) {
        elapsed = waveletDisk();
        newResult(10,elapsed);
        }

        // 11 - Adaptive K-D tree on disk:
    if ( mask & 0x0400 ) {
        elapsed = akdTree();
        newResult(11,elapsed);
        }

    writeData();
    return 0;
}
